@@ -1,11 +1,8 @@
 # frozen_string_literal: true
 
-require 'minitest/global_expectations/autorun'
-require 'stringio'
+require_relative 'helper'
 require 'cgi'
-require 'rack/request'
-require 'rack/mock'
-require 'rack/multipart'
+require 'forwardable'
 require 'securerandom'
 
 class RackRequestTest < Minitest::Spec
@@ -117,24 +114,37 @@ class RackRequestTest < Minitest::Spec
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "www2.example.org")
     req.host.must_equal "www2.example.org"
+    req.hostname.must_equal "www2.example.org"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "123foo.example.com")
+    req.host.must_equal "123foo.example.com"
+    req.hostname.must_equal "123foo.example.com"
 
     req = make_request \
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.host.must_equal "example.org"
+    req.hostname.must_equal "example.org"
 
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org:9292")
     req.host.must_equal "example.org"
+    req.hostname.must_equal "example.org"
 
-    env = Rack::MockRequest.env_for("/", "SERVER_ADDR" => "192.168.1.1", "SERVER_PORT" => "9292")
-    env.delete("SERVER_NAME")
-    req = make_request(env)
-    req.host.must_equal "192.168.1.1"
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[2001:db8:cafe::17]:47011")
+    req.host.must_equal "[2001:db8:cafe::17]"
+    req.hostname.must_equal "2001:db8:cafe::17"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "2001:db8:cafe::17")
+    req.host.must_equal "[2001:db8:cafe::17]"
+    req.hostname.must_equal "2001:db8:cafe::17"
 
     env = Rack::MockRequest.env_for("/")
     env.delete("SERVER_NAME")
     req = make_request(env)
-    req.host.must_equal ""
+    req.host.must_be_nil
   end
 
   it "figure out the correct port" do
@@ -154,6 +164,14 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org:9292")
     req.port.must_equal 9292
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[2001:db8:cafe::17]:47011")
+    req.port.must_equal 47011
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "2001:db8:cafe::17")
+    req.port.must_equal 80
+
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org")
     req.port.must_equal 80
@@ -162,7 +180,7 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "HTTP_X_FORWARDED_SSL" => "on")
     req.port.must_equal 443
 
-     req = make_request \
+    req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "HTTP_X_FORWARDED_PROTO" => "https")
     req.port.must_equal 443
 
@@ -200,10 +218,22 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.host_with_port.must_equal "example.org:9292"
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org")
+    req.host_with_port.must_equal "example.org"
+
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org:9292")
     req.host_with_port.must_equal "example.org:9292"
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[2001:db8:cafe::17]:47011")
+    req.host_with_port.must_equal "[2001:db8:cafe::17]:47011"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "2001:db8:cafe::17")
+    req.host_with_port.must_equal "[2001:db8:cafe::17]"
+
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "SERVER_PORT" => "9393")
     req.host_with_port.must_equal "example.org"
@@ -262,7 +292,7 @@ class RackRequestTest < Minitest::Spec
     old, Rack::Utils.key_space_limit = Rack::Utils.key_space_limit, 1
     begin
       req = make_request(env)
-      lambda { req.GET }.must_raise RangeError
+      lambda { req.GET }.must_raise Rack::QueryParser::ParamsTooDeepError
     ensure
       Rack::Utils.key_space_limit = old
     end
@@ -276,12 +306,37 @@ class RackRequestTest < Minitest::Spec
     begin
       exp = { "foo" => { "bar" => { "baz" => { "qux" => "1" } } } }
       make_request(nested_query).GET.must_equal exp
-      lambda { make_request(plain_query).GET  }.must_raise RangeError
+      lambda { make_request(plain_query).GET  }.must_raise Rack::QueryParser::ParamsTooDeepError
     ensure
       Rack::Utils.key_space_limit = old
     end
   end
 
+  it "limit the allowed parameter depth when parsing parameters" do
+    env = Rack::MockRequest.env_for("/?a#{'[a]' * 110}=b")
+    req = make_request(env)
+    lambda { req.GET }.must_raise Rack::QueryParser::ParamsTooDeepError
+
+    env = Rack::MockRequest.env_for("/?a#{'[a]' * 90}=b")
+    req = make_request(env)
+    params = req.GET
+    90.times { params = params['a'] }
+    params['a'].must_equal 'b'
+
+    old, Rack::Utils.param_depth_limit = Rack::Utils.param_depth_limit, 3
+    begin
+      env = Rack::MockRequest.env_for("/?a[a][a]=b")
+      req = make_request(env)
+      req.GET['a']['a']['a'].must_equal 'b'
+
+      env = Rack::MockRequest.env_for("/?a[a][a][a]=b")
+      req = make_request(env)
+      lambda { make_request(env).GET  }.must_raise Rack::QueryParser::ParamsTooDeepError
+    ensure
+      Rack::Utils.param_depth_limit = old
+    end
+  end
+
   it "not unify GET and POST when calling params" do
     mr = Rack::MockRequest.env_for("/?foo=quux",
       "REQUEST_METHOD" => 'POST',
@@ -361,7 +416,7 @@ class RackRequestTest < Minitest::Spec
     old, Rack::Utils.key_space_limit = Rack::Utils.key_space_limit, 1
     begin
       req = make_request(env)
-      lambda { req.POST }.must_raise RangeError
+      lambda { req.POST }.must_raise Rack::QueryParser::ParamsTooDeepError
     ensure
       Rack::Utils.key_space_limit = old
     end
@@ -388,6 +443,7 @@ class RackRequestTest < Minitest::Spec
     req.content_type.must_equal 'text/plain;charset=utf-8'
     req.media_type.must_equal 'text/plain'
     req.media_type_params['charset'].must_equal 'utf-8'
+    req.content_charset.must_equal 'utf-8'
     req.POST.must_be :empty?
     req.params.must_equal "foo" => "quux"
     req.body.read.must_equal "foo=bar&quux=bla"
@@ -439,6 +495,20 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("?foo=quux")
     req['foo'].must_equal 'quux'
     req[:foo].must_equal 'quux'
+
+    next if self.class == TestProxyRequest
+    verbose = $VERBOSE
+    warn_arg = nil
+    req.define_singleton_method(:warn) do |arg|
+      warn_arg = arg
+    end
+    begin
+      $VERBOSE = true
+      req['foo'].must_equal 'quux'
+      warn_arg.must_equal "Request#[] is deprecated and will be removed in a future version of Rack. Please use request.params[] instead"
+    ensure
+      $VERBOSE = verbose
+    end
   end
 
   it "set value to key on params with #[]=" do
@@ -461,6 +531,20 @@ class RackRequestTest < Minitest::Spec
     req.params.must_equal 'foo' => 'jaz'
     req['foo'].must_equal 'jaz'
     req[:foo].must_equal 'jaz'
+
+    verbose = $VERBOSE
+    warn_arg = nil
+    req.define_singleton_method(:warn) do |arg|
+      warn_arg = arg
+    end
+    begin
+      $VERBOSE = true
+      req['foo'] = 'quux'
+      warn_arg.must_equal "Request#[]= is deprecated and will be removed in a future version of Rack. Please use request.params[]= instead"
+      req.params['foo'].must_equal 'quux'
+    ensure
+      $VERBOSE = verbose
+    end
   end
 
   it "return values for the keys in the order given from values_at" do
@@ -549,6 +633,10 @@ class RackRequestTest < Minitest::Spec
     request.scheme.must_equal "https"
     request.must_be :ssl?
 
+    request = make_request(Rack::MockRequest.env_for("/", 'rack.url_scheme' => 'wss'))
+    request.scheme.must_equal "wss"
+    request.must_be :ssl?
+
     request = make_request(Rack::MockRequest.env_for("/", 'HTTP_HOST' => 'www.example.org:8080'))
     request.scheme.must_equal "http"
     request.wont_be :ssl?
@@ -583,7 +671,6 @@ class RackRequestTest < Minitest::Spec
     req = make_request \
       Rack::MockRequest.env_for("", "HTTP_COOKIE" => "foo=bar;quux=h&m")
     req.cookies.must_equal "foo" => "bar", "quux" => "h&m"
-    req.cookies.must_equal "foo" => "bar", "quux" => "h&m"
     req.delete_header("HTTP_COOKIE")
     req.cookies.must_equal({})
   end
@@ -913,7 +1000,7 @@ EOF
     f[:tempfile].size.must_equal 76
   end
 
-  it "MultipartPartLimitError when request has too many multipart parts if limit set" do
+  it "MultipartPartLimitError when request has too many multipart file parts if limit set" do
     begin
       data = 10000.times.map { "--AaB03x\r\nContent-Type: text/plain\r\nContent-Disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
       data += "--AaB03x--\r"
@@ -929,6 +1016,22 @@ EOF
     end
   end
 
+  it "MultipartPartLimitError when request has too many multipart total parts if limit set" do
+    begin
+      data = 10000.times.map { "--AaB03x\r\ncontent-type: text/plain\r\ncontent-disposition: attachment; name=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
+      data += "--AaB03x--\r"
+
+      options = {
+        "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
+        "CONTENT_LENGTH" => data.length.to_s,
+        :input => StringIO.new(data)
+      }
+
+      request = make_request Rack::MockRequest.env_for("/", options)
+      lambda { request.POST }.must_raise Rack::Multipart::MultipartTotalPartLimitError
+    end
+  end
+
   it 'closes tempfiles it created in the case of too many created' do
     begin
       data = 10000.times.map { "--AaB03x\r\nContent-Type: text/plain\r\nContent-Disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
@@ -1188,6 +1291,12 @@ EOF
 
     res = mock.get '/', 'REMOTE_ADDR' => '1.2.3.4,3.4.5.6'
     res.body.must_equal '1.2.3.4'
+
+    res = mock.get '/', 'REMOTE_ADDR' => '127.0.0.1'
+    res.body.must_equal '127.0.0.1'
+
+    res = mock.get '/', 'REMOTE_ADDR' => '127.0.0.1,127.0.0.1'
+    res.body.must_equal '127.0.0.1'
   end
 
   it 'deals with proxies' do
